#include "extended_type_info.hpp"
#include "type_info_parser.hpp"
#include <stack>
#include <exception>
#include <boost/spirit.hpp>

namespace boost
{
namespace detail
{
namespace dll
{
std::map<std::string, extended_type_info::type_t> extended_type_info::base_type_map_;
bool extended_type_info::base_type_map_initialized_ = false;

extended_type_info::extended_type_info(const std::type_info &info)
{
	if(!base_type_map_initialized_)
	{
		initialize_base_types();
		base_type_map_initialized_ = true;
	}

	*this = parse(info.name());
}

extended_type_info::extended_type_info(const extended_type_info &info)
	: name_(info.name_),
	  type_(info.type_),
	  base_type_(info.base_type_),
	  type_name_(info.type_name_),
	  namespaces_(info.namespaces_),
	  template_arguments_(info.template_arguments_),
	  return_type_(info.return_type_),
	  function_parameters_(info.function_parameters_)
{
}

extended_type_info::extended_type_info(const std::string &str)
{
	if(!base_type_map_initialized_)
	{
		initialize_base_types();
		base_type_map_initialized_ = true;
	}

	*this = parse(str);
}

extended_type_info::extended_type_info()
{
	type_ = void_type;
	type_name_ = base_type_to_string(type_);

	build_type_name();
	build_name();
}

extended_type_info::~extended_type_info()
{
}

const std::vector< boost::shared_ptr<extended_type_info> > &boost::detail::dll::extended_type_info::template_arguments() const
{
	return template_arguments_;
}

extended_type_info extended_type_info::parse(const std::string &str)
{
	type_info_parser::parsing_context context;
	type_info_grammar grammar(context);
	boost::spirit::parse_info<char const *> info = boost::spirit::parse(str.c_str(), grammar, boost::spirit::space_p);

	//
	// FIXME: check only in debug builds
	if(context.info_stack.size() != 1 || !info.full)
	{
		std::cerr << "extended_type_info::parse: damn: " << info.stop << std::endl;
		throw std::runtime_error("malformed type string");
	}

	context.info_stack.top().build_type_name();
	context.info_stack.top().build_name();

	return context.info_stack.top();
}

void extended_type_info::build_name()
{
	if(type_ & (class_type | struct_type | union_type | enum_type))
	{
		name_ = base_type_to_string(type_) + " " + type_name_;
		
		//
		// append template parameters
		if(template_arguments_.size())
		{
			name_ += "< ";
			for(std::vector< boost::shared_ptr<extended_type_info> >::iterator i = template_arguments_.begin(); i != template_arguments_.end(); ++i)
			{
				(*i)->build_name();
				if(i != template_arguments_.begin())
					name_ += ", ";
				name_ += (*i)->name_;
			}
			name_ += " >";
		}
		return;
	}
	else if(type_ & (pointer_type | reference_type))
	{
		base_type_->build_name();

		boost::shared_ptr<extended_type_info> info = base_type_;
		size_t num_ptrs = 1, num_refs = 0;
		while(info->type_ & (pointer_type | reference_type))
		{
			(info->type_ == pointer_type) ? ++num_ptrs : ++num_refs;
			info = info->base_type_;
		}

		if(info->type_ & function_type)
		{ 
			//
			// hacky: since base is a function we need to handle the type name
			//		  a bit different
			name_ = build_function_name_string(info.get(), num_ptrs, num_refs);
		}
		else
		{
			name_ = base_type_->name_;

			if(!(base_type_->type_ & (pointer_type | reference_type)))
				name_ += ' ';

			if(type_ == pointer_type)
				name_ += '*';
			else
				name_ += '&';
		}
		return;
	}
	else if(type_ == function_type)
	{
		return_type_->build_name();
		for(std::vector< boost::shared_ptr<extended_type_info> >::iterator i = function_parameters_.begin(); i != function_parameters_.end(); ++i)
			(*i)->build_name();
		name_ = build_function_name_string(this);
		return;
	}

	name_ = base_type_to_string(type_);
}

void extended_type_info::build_type_name()
{
	if(type_ & (class_type | struct_type | union_type | enum_type))
		return; // already set while parsing

	if(type_ & (pointer_type | reference_type))
	{
		base_type_->build_type_name();

		boost::shared_ptr<extended_type_info> info = base_type_;
		size_t num_ptrs = 1, num_refs = 0;
		while(info->type_ & (pointer_type | reference_type))
		{
			(info->type_ == pointer_type) ? ++num_ptrs : ++num_refs;
			info = info->base_type_;
		}

		if(info->type_ & function_type)
		{ 
			//
			// hacky: since base is a function we need to handle the type name
			//		  a bit different
			type_name_ = build_function_type_name_string(info.get(), num_ptrs, num_refs);
		}
		else
		{
			type_name_ = base_type_->type_name_;

			if(!(base_type_->type_ & (pointer_type | reference_type)))
				type_name_ += ' ';

			if(type_ == pointer_type)
				type_name_ += '*';
			else
				type_name_ += '&';
		}
		return;
	}
	else if(type_ == function_type)
	{
		return_type_->build_type_name();
		for(std::vector< boost::shared_ptr<extended_type_info> >::iterator i = function_parameters_.begin(); i != function_parameters_.end(); ++i)
			(*i)->build_type_name();
		type_name_ = build_function_type_name_string(this);
		return;
	}

	type_name_ = base_type_to_string(type_);
}

std::string extended_type_info::build_function_name_string(const extended_type_info *fn,
														   size_t num_ptrs,
														   size_t num_refs)
{
	std::string ret;

	ret = fn->return_type_->name_;
	if(fn->namespaces_.size() || num_ptrs + num_refs != 0)
	{
		ret += std::string(" (") + build_namespace_string(fn);
		if(fn->namespaces_.size() && num_ptrs + num_refs != 0)
			ret += ' ';

		for(size_t a = 0; a < num_ptrs; ++a)
			ret += '*';

		for(size_t a = 0; a < num_refs; ++a)
			ret += '&';
		ret += ')';
	}

	ret += '(';
	if(!fn->function_parameters_.size())
		ret += "void";
	else
	{
		for(std::vector< boost::shared_ptr<extended_type_info> >::const_iterator i = fn->function_parameters_.begin(); i != fn->function_parameters_.end(); ++i)
		{
			if(i != fn->function_parameters_.begin())
				ret += ", ";
			ret += (*i)->name_;
		}
	}
	ret += ')';

	return ret;
}

std::string extended_type_info::build_function_type_name_string(const extended_type_info *fn,
																size_t num_ptrs,
																size_t num_refs)
{
	std::string ret;

	ret = fn->return_type_->type_name_;
	if(fn->namespaces_.size() || num_ptrs + num_refs != 0)
	{
		ret += std::string(" (") + build_namespace_string(fn);
		if(fn->namespaces_.size() && num_ptrs + num_refs != 0)
			ret += ' ';

		for(size_t a = 0; a < num_ptrs; ++a)
			ret += '*';

		for(size_t a = 0; a < num_refs; ++a)
			ret += '&';
		ret += ')';
	}

	ret += '(';
	if(!fn->function_parameters_.size())
		ret += "void";
	else
	{
		for(std::vector< boost::shared_ptr<extended_type_info> >::const_iterator i = fn->function_parameters_.begin(); i != fn->function_parameters_.end(); ++i)
		{
			if(i != fn->function_parameters_.begin())
				ret += ", ";
			ret += (*i)->type_name_;
		}
	}
	ret += ')';

	return ret;
}

std::string extended_type_info::build_namespace_string(const extended_type_info *info)
{
	if(!info->namespaces_.size())
		return "";

	std::string ret;
	for(std::vector<std::string>::const_iterator i = info->namespaces_.begin(); i != info->namespaces_.end(); ++i)
		ret += *i + "::";
	return ret;
}

extended_type_info::type_t extended_type_info::string_to_base_type(const std::string &type_str)
{
	std::map<std::string, type_t>::const_iterator i = base_type_map_.find(type_str);
	
	//
	// FIXME: only in debug builds
	if(i == base_type_map_.end())
	{
		std::cerr << "extended_type_info::string_to_base_type(" << type_str << "): no such type" << std::endl;
		throw std::runtime_error("malformed type string");
	}

	return (*i).second;
}

std::string extended_type_info::base_type_to_string(extended_type_info::type_t type)
{
	//
	// FIXME: slow

	for(std::map<std::string, type_t>::const_iterator i = base_type_map_.begin(); i != base_type_map_.end(); ++i)
		if((*i).second == type) // note: not & since we need exact match
			return (*i).first;

	//
	// FIXME: only in debug builds
	throw std::runtime_error("malformed type string");
	return ""; // to disable warning
}

void extended_type_info::initialize_base_types()
{
	base_type_map_["void"] = void_type;

	base_type_map_["char"] = signed_char_type;
	base_type_map_["signed char"] = signed_char_type;
	base_type_map_["unsigned char"] = unsigned_char_type;
	base_type_map_["short"] = signed_short_type;
	base_type_map_["signed short"] = signed_short_type;
	base_type_map_["unsigned short"] = unsigned_short_type;
	base_type_map_["int"] = signed_int_type;
	base_type_map_["signed int"] = signed_int_type;
	base_type_map_["unsigned int"] = unsigned_int_type;
	base_type_map_["long"] = signed_long_type;
	base_type_map_["signed long"] = signed_long_type;
	base_type_map_["unsigned long"] = unsigned_long_type;

	base_type_map_["float"] = float_type;
	base_type_map_["double"] = double_type;

	base_type_map_["class"] = class_type;
	base_type_map_["struct"] = struct_type;
	base_type_map_["union"] = union_type;
}

}; // end namespace dll
}; // end namespace detail
}; // end namespace boost