/*
	extended_type_info.hpp
*/
#ifndef BOOST_DLL_EXTENDED_TYPE_INFO_HPP
#define BOOST_DLL_EXTENDED_TYPE_INFO_HPP

#include <map>
#include <string>
#include <vector>
#include <typeinfo>
#include <boost/shared_ptr.hpp>

namespace boost
{
	namespace detail
	{
		namespace dll
		{
			class extended_type_info;

			template<typename T>
			extended_type_info type_id();

			//
			// really ugly, need prototypes for friend declarations
			namespace type_info_parser
			{
				struct parsing_context;
				struct push_basic_type;
				struct set_ptr_type;
				struct set_ref_type;
				struct set_cs_type_name;
				struct add_namespace;
				struct add_template_argument;
				struct set_return_type;
				struct add_function_parameter;
			};

			//
			// FIXME: the class will be heavily copied while parsing, maybe
			//		  we need an implicit shared implementation to speed things
			//		  up
			class extended_type_info
			{
				friend struct type_info_parser::parsing_context;
				friend struct type_info_parser::push_basic_type;
				friend struct type_info_parser::set_ptr_type;
				friend struct type_info_parser::set_ref_type;
				friend struct type_info_parser::set_cs_type_name;
				friend struct type_info_parser::add_namespace;
				friend struct type_info_parser::add_template_argument;
				friend struct type_info_parser::set_return_type;
				friend struct type_info_parser::add_function_parameter;

			public:
				enum type_t
				{
					void_type = 1 << 0,
					signed_char_type = 1 << 1,
					unsigned_char_type = 1 << 2,
					signed_short_type = 1 << 3,
					unsigned_short_type = 1 << 4,
					signed_int_type = 1 << 5,
					unsigned_int_type = 1 << 6,
					signed_long_type = 1 << 7,
					unsigned_long_type = 1 << 8,
					float_type = 1 << 9,
					double_type = 1 << 10,

					pointer_type = 1 << 12,
					reference_type = 1 << 13,
					
					union_type = 1 << 15,
					class_type = 1 << 16,
					struct_type = 1 << 17,
					enum_type = 1 << 18,

					function_type = 1 << 20,

					builtin_type_mask = void_type |
										signed_char_type |
										unsigned_char_type |
										signed_short_type |
										unsigned_short_type |
										signed_int_type |
										unsigned_int_type |
										signed_long_type |
										unsigned_long_type |
										float_type |
										double_type
				};

				extended_type_info(const std::type_info &info);
				extended_type_info(const extended_type_info &info);
				~extended_type_info();

			protected:
				extended_type_info(); // constructs type info for 'void'
				extended_type_info(const std::string &str);

			public:
				type_t type() const;
				
				//
				// returns the type string, same as std::type_info::name()
				const std::string &name() const;

				//
				// the string representing the base
				// note: only set for pointers or references otherwise
				//		 returns *this
				const extended_type_info &base_type() const;

				//
				// returns the type name, e.g. "int", "char" for PODS
				// and the class/struct/union name for classes/structs/unions
				// note: the typename does not contain any namespaces
				const std::string &type_name() const;

				//
				// returns the type name with prepended namespace string if
				// with_namespaces is true, otherwise same as type_name()
				std::string type_name(bool with_namespaces) const;

				//
				// returns vector of namespaces
				const std::vector<std::string> &namespaces() const;

				//
				// returns vector of template arguments
				const std::vector< boost::shared_ptr<extended_type_info> > &template_arguments() const;

				extended_type_info &operator=(const extended_type_info &info);

			private:
				static extended_type_info parse(const std::string &str);
				
				//
				// build a string similar to that of returned by type_info.name()
				// note: we do not take the string of type_info since the parser parses
				//		 the compiler dependent string and produces a compiler independent
				//		 string
				void build_name();

				//
				// build type which reflects type_ and which is useful to the user
				// e.g. int **
				// note: this is only called for non class/struct/union/enum types
				void build_type_name();

				static std::string build_function_name_string(const extended_type_info *fn,
															  size_t num_ptrs = 0,
															  size_t num_refs = 0);
				static std::string build_function_type_name_string(const extended_type_info *fn,
																   size_t num_ptrs = 0,
																   size_t num_refs = 0);

				//
				// returns a string which represents the namespaces or "" if
				// no namespaces specified
				static std::string build_namespace_string(const extended_type_info *info);
				
				//
				// note: pointers/references are not supported
				static type_t string_to_base_type(const std::string &type_str);
				static std::string base_type_to_string(type_t type);

				static void initialize_base_types();

			private:
				std::string							  name_;
				type_t								  type_;
				boost::shared_ptr<extended_type_info> base_type_;
				std::string							  type_name_;

				std::vector<std::string>							 namespaces_;
				std::vector< boost::shared_ptr<extended_type_info> > template_arguments_;

				//
				// FIXME: these data members are not needed for non function types, how
				//		  can we safe this space?! (union works only for PODs)
				boost::shared_ptr<extended_type_info>				 return_type_;
				std::vector< boost::shared_ptr<extended_type_info> > function_parameters_;

				static std::map<std::string, type_t> base_type_map_;
				static bool						   base_type_map_initialized_;
			};

		}; // end namespace dll
	}; // end namespace detail
}; // end namespace boost

// -- inline definitions --
template<typename T>
inline boost::detail::dll::extended_type_info boost::detail::dll::type_id()
{
	return extended_type_info(typeid(T));	
}

//
// class *** boost::detail::dll::extended_type_info ***
// 
inline boost::detail::dll::extended_type_info::type_t boost::detail::dll::extended_type_info::type() const
{
	return type_;
}

inline const std::string &boost::detail::dll::extended_type_info::name() const
{
	return name_;
}

inline const std::string &boost::detail::dll::extended_type_info::type_name() const
{
	return type_name_;
}

inline std::string boost::detail::dll::extended_type_info::type_name(bool with_namespaces) const
{
	if(!with_namespaces)
		return type_name_;

	return build_namespace_string(this) + type_name_;
}

inline const std::vector<std::string> &boost::detail::dll::extended_type_info::namespaces() const
{
	return namespaces_;
}

/*inline const std::vector< boost::shared_ptr<extended_type_info> > &boost::detail::dll::extended_type_info::template_arguments() const
{
	// HACK: cannot be inline, MSVC bug?
	return template_arguments_;
}*/

inline const boost::detail::dll::extended_type_info &boost::detail::dll::extended_type_info::base_type() const
{
	if(!base_type_)
		return *this;
	return *base_type_.get();
}

inline boost::detail::dll::extended_type_info &boost::detail::dll::extended_type_info::operator =(const boost::detail::dll::extended_type_info &info)
{
	name_ = info.name_;
	type_ = info.type_;
	base_type_ = info.base_type_;
	type_name_ = info.type_name_;
	namespaces_ = info.namespaces_;
	template_arguments_ = info.template_arguments_;
	return_type_ = info.return_type_;
	function_parameters_ = info.function_parameters_;
	return *this;
}

inline bool operator ==(const boost::detail::dll::extended_type_info &info1, const boost::detail::dll::extended_type_info &info2)
{
	//
	// as every information is encoded in the name we can compare that, should
	// be much faster
	return info1.name() == info2.name();
}

inline bool operator !=(const boost::detail::dll::extended_type_info &info1, const boost::detail::dll::extended_type_info &info2)
{
	return !(info1 == info2);
}

#endif // BOOST_DLL_EXTENDED_TYPE_INFO_HPP