/*
	type_info_parser.hpp

	parser for strings returned by type_info::name()
*/
#ifndef BOOST_DLL_TYPE_INFO_PARSER_HPP
#define BOOST_DLL_TYPE_INFO_PARSER_HPP

#include "extended_type_info.hpp"
#include <map> // for std::pair
#include <stack>
#include <boost/spirit.hpp>

namespace boost
{
	namespace detail
	{
		namespace dll
		{
			//
			// this namespace collects all semantic actions used
			// by boost::spirit
			namespace type_info_parser
			{
				struct parsing_context
				{
					parsing_context()
					{
					}

					std::stack<extended_type_info> info_stack;
					//
					// we need to keep track of the number of references for
					// function pointers (since we need to parse the complete type
					// before we can set it as the base_type_ but we read the * or &
					// before the complete type is build)
					std::stack< std::pair<size_t, size_t> > post_fn_ptr_refs;
				};

				struct push_basic_type
				{
					push_basic_type(parsing_context &c) : context(c) {}

					void operator()(char const *first, char const *last) const
					{
						//
						// push new basic type on stack
						context.info_stack.push(extended_type_info());
						context.info_stack.top().type_ = extended_type_info::string_to_base_type(std::string(first, last));
					}
					
					parsing_context &context;
				};

				struct set_ptr_type
				{
					set_ptr_type(parsing_context &c) : context(c) {}

					void operator()(char c) const
					{
						extended_type_info base = context.info_stack.top();
						context.info_stack.pop();
						
						extended_type_info ptr_info;
						ptr_info.type_ = extended_type_info::pointer_type;
						ptr_info.base_type_ = boost::shared_ptr<extended_type_info>(new extended_type_info(base));
						context.info_stack.push(ptr_info);
					}
					
					parsing_context &context;
				};

				struct set_ref_type
				{
					set_ref_type(parsing_context &c) : context(c) {}

					void operator()(char c) const
					{
						extended_type_info base = context.info_stack.top();
						context.info_stack.pop();
						
						extended_type_info ref_info;
						ref_info.type_ = extended_type_info::reference_type;
						ref_info.base_type_ = boost::shared_ptr<extended_type_info>(new extended_type_info(base));
						context.info_stack.push(ref_info);
					}
					
					parsing_context &context;
				};

				struct set_cs_type_name
				{
					set_cs_type_name(parsing_context &c) : context(c) {}

					void operator()(char const *first, char const *last) const
					{
						context.info_stack.top().type_name_ = std::string(first, last);
					}
					
					parsing_context &context;
				};

				struct add_namespace
				{
					add_namespace(parsing_context &c) : context(c) {}

					void operator()(char const *first, char const *last) const
					{
						std::string ns(first, last - 2); // cut of ::
						context.info_stack.top().namespaces_.push_back(ns);
					}
					
					parsing_context &context;
				};

				struct add_template_argument
				{
					add_template_argument(parsing_context &c) : context(c) {}

					void operator()(char const *first, char const *last) const
					{
						//
						// note: the template argument is on top of the stack, pop it and
						//		 add it to the info object which is now on top
						extended_type_info argument = context.info_stack.top();
						context.info_stack.pop();
						context.info_stack.top().template_arguments_.push_back(boost::shared_ptr<extended_type_info>(new extended_type_info(argument)));
					}
					
					parsing_context &context;
				};

				struct set_return_type
				{
					set_return_type(parsing_context &c) : context(c) {}

					void operator()(char c) const
					{
						//
						// note: the return type is on top of the stack, pop it and push
						//		 a function object
						extended_type_info type = context.info_stack.top();
						context.info_stack.pop();
						context.info_stack.push(extended_type_info());
						context.info_stack.top().type_ = extended_type_info::function_type;
						context.info_stack.top().return_type_ = boost::shared_ptr<extended_type_info>(new extended_type_info(type));

						//
						// push referencing object
						context.post_fn_ptr_refs.push(std::pair<size_t, size_t>(0, 0));
					}	
					
					parsing_context &context;
				};

				struct add_function_parameter
				{
					add_function_parameter(parsing_context &c) : context(c) {}

					void operator()(char const *first, char const *last) const
					{
						//
						// note: the function parameter is on top of the stack, pop it and
						//		 add it to the info object which is now on top
						extended_type_info parameter = context.info_stack.top();
						context.info_stack.pop();
						context.info_stack.top().function_parameters_.push_back(boost::shared_ptr<extended_type_info>(new extended_type_info(parameter)));
					}
					
					parsing_context &context;
				};

				struct add_post_ptr
				{
					add_post_ptr(parsing_context &c) : context(c) {}

					void operator()(char c) const
					{
						++context.post_fn_ptr_refs.top().first;
					}
					
					parsing_context &context;
				};

				struct add_post_ref
				{
					add_post_ref(parsing_context &c) : context(c) {}

					void operator()(char c) const
					{
						++context.post_fn_ptr_refs.top().second;
					}
					
					parsing_context &context;
				};

				struct post_adjust
				{
					post_adjust(parsing_context &c) : context(c) {}

					void operator()(char const *first, char const *last) const
					{
						//
						// first push num_post_ptrs
						for(size_t a = 0; a < context.post_fn_ptr_refs.top().first; ++a)
						{
							//
							// BIIIIG HACK: call functor directly (since it does not interpret the
							//				char paramater)
							set_ptr_type hack(context);
							hack('H'); // H for HACK
						}

						//
						// then push num_post_regs
						for(size_t a = 0; a < context.post_fn_ptr_refs.top().second; ++a)
						{
							//
							// BIIIIG HACK: call functor directly (since it does not interpret the
							//				char oarameter)
							set_ref_type hack(context);
							hack('H'); // H for HACK
						}

						context.post_fn_ptr_refs.pop();
					}
					
					parsing_context &context;
				};
			}; // end namespace type_info_parser

			struct type_info_grammar : public boost::spirit::grammar<type_info_grammar>
			{
				type_info_grammar(type_info_parser::parsing_context &c) : context(c) {}

				type_info_parser::parsing_context &context;

				template <typename ScannerT>
				struct definition
				{
					boost::spirit::rule<ScannerT> type;
					boost::spirit::rule<ScannerT> primitive;
					boost::spirit::rule<ScannerT> pod_primitive;
					boost::spirit::rule<ScannerT> cs_primitive;
					boost::spirit::rule<ScannerT> enum_primitive;
					boost::spirit::rule<ScannerT> union_primitive;
					boost::spirit::rule<ScannerT> template_list;
					boost::spirit::rule<ScannerT> parameter_list;
					boost::spirit::rule<ScannerT> namespace_list;
					boost::spirit::rule<ScannerT> calling_convention;
					boost::spirit::rule<ScannerT> basic_type;
					boost::spirit::rule<ScannerT> cs; // class, struct
					boost::spirit::rule<ScannerT> identifier;

					definition(type_info_grammar const& self)
					{
						type = primitive >> !(
								boost::spirit::ch_p('(')[type_info_parser::set_return_type(self.context)] >>
								calling_convention >>
								namespace_list >>
								+(boost::spirit::ch_p('*')[type_info_parser::add_post_ptr(self.context)] |
								boost::spirit::ch_p('&')[type_info_parser::add_post_ref(self.context)]) >>
								boost::spirit::ch_p(')') >>
								'(' >> parameter_list[type_info_parser::post_adjust(self.context)] >> ')'
							);
						primitive = (pod_primitive | cs_primitive | enum_primitive | union_primitive) >>
									*(boost::spirit::ch_p('*')[type_info_parser::set_ptr_type(self.context)] |
									boost::spirit::ch_p('&')[type_info_parser::set_ref_type(self.context)]);
						pod_primitive = basic_type[type_info_parser::push_basic_type(self.context)];
						cs_primitive = cs[type_info_parser::push_basic_type(self.context)] >>
									namespace_list >>
									identifier[type_info_parser::set_cs_type_name(self.context)] >>
									!(boost::spirit::ch_p('<') >> template_list >> boost::spirit::ch_p('>'));
						enum_primitive = boost::spirit::str_p("enum") >> identifier;   // FIXME: need semantic actions
						union_primitive = boost::spirit::str_p("union") >> identifier; // FIXME: need semantic actions

						template_list = boost::spirit::list_p(type[type_info_parser::add_template_argument(self.context)], boost::spirit::ch_p(','));
						namespace_list = *((identifier >> "::")[type_info_parser::add_namespace(self.context)]);
						parameter_list = boost::spirit::list_p(type[type_info_parser::add_function_parameter(self.context)], boost::spirit::ch_p(','));
						calling_convention = boost::spirit::str_p("__cdecl") | "__thiscall";

						basic_type = boost::spirit::str_p("void") |
									"char" |
									boost::spirit::str_p("signed") >> "char" |
									boost::spirit::str_p("unsigned") >> "char" |
									"short" |
									boost::spirit::str_p("signed") >> "short" |
									boost::spirit::str_p("unsigned") >> "short" |
									"int" |
									boost::spirit::str_p("signed") >> "int" |
									boost::spirit::str_p("unsigned") >> "int" |
									"long" |
									boost::spirit::str_p("signed") >> "long" |
									boost::spirit::str_p("unsigned") >> "long" |
									"float" | 
									"double";
						cs = boost::spirit::str_p("class") | boost::spirit::str_p("struct");

						identifier = boost::spirit::lexeme_d[boost::spirit::alpha_p >> *('_' | boost::spirit::alnum_p)];
					}

					boost::spirit::rule<ScannerT> const& start() const { return type; }
				};
			};
		};
	};
};

#endif // BOOST_DLL_TYPE_INFO_PARSER_HPP